/*	$LAAS$ */
/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#include <portLib.h>
#include <errnoLib.h>

#include <neobotix/pscan.h>
#include <neobotix/neo.h>
#include <neobotix/canGyro.h>

#include "canGyroCmd.h"
#include "canIds.h"

/*----------------------------------------------------------------------*/
/*
 * Initialize communication with the gyroscope 
 */
CAN_GYRO_STR *
canGyroInit(CAN_DEV *dev)
{
	CAN_GYRO_STR *g;

	g = (CAN_GYRO_STR *)malloc(sizeof(CAN_GYRO_STR));
	if (g == NULL) {
		errnoSet(S_canGyro_NO_MEMORY);
		return NULL;
	}
	g->mutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	if (g->mutex == NULL) {
		free(g);
		errnoSet(S_canGyro_NO_MEMORY);
		return NULL;
	}
	pthread_mutex_init(g->mutex, NULL);
	g->dev = dev;
	g->scale = 1./520.5;
	g->timeout = 100000;		/* 100ms */
	g->recMsgId = GYROBOARD_RX_ID;
	return g;
}

/*----------------------------------------------------------------------*/
/*
 * End communication with the gyroscope
 */
void
canGyroEnd(CAN_GYRO_STR *gyro)
{
	pthread_mutex_destroy(gyro->mutex);
	free(gyro);
}

/*----------------------------------------------------------------------*/
/*
 * Send the 'connect' message to the gyroscope 
 * (needed before any other communication can take place)
 */
void
canGyroRequestConnect(NEO_STR *neo)
{
	CAN_GYRO_STR *gyro = neo->gyro;
	CAN_MSG m;

	canMsgSet(&m, GYROBOARD_TX_ID, 
	    CMD_GYROBOARD_CONNECT, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(gyro->dev, &m);

}
       
/*----------------------------------------------------------------------*/
/*
 * Send 2 messages requesting the gyroscope's data
 */
void
canGyroRequestAngle(NEO_STR *neo)
{
	CAN_GYRO_STR *gyro = neo->gyro;
	CAN_MSG m;
	
	canMsgSet(&m, GYROBOARD_TX_ID, 
	    CMD_GYROBOARD_GETGYRO, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(gyro->dev, &m);
	
	canMsgSet(&m, GYROBOARD_TX_ID,
	    CMD_GYROBOARD_GETDATA, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(gyro->dev, &m);

}

/*----------------------------------------------------------------------*/
/*
 * Set the gyroscope to zero 
 */
void
canGyroZero(NEO_STR *neo)
{
	CAN_GYRO_STR *gyro = neo->gyro;
	CAN_MSG m;

	canMsgSet(&m, GYROBOARD_TX_ID, 
	    CMD_GYROBOARD_ZEROGYRO, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(gyro->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Request the acceleration data (?)
 */
void
canGyroRequestAccel(NEO_STR *neo)
{
	CAN_GYRO_STR *gyro = neo->gyro;
	CAN_MSG m;
	
	canMsgSet(&m, GYROBOARD_TX_ID, 
	    CMD_GYROBOARD_GETDATA, 0, 0, 0, 0, 0, 0, 0);
	pscanTransmitMsg(gyro->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Decode CAN messages sent by the gyroscope 
 */
BOOL
canGyroDecode(CAN_GYRO_STR *gyro, CAN_MSG *m)
{
	int cmd = canGetCmd(m);

	if (m->id != gyro->recMsgId) {
		
		return FALSE;
	}
	pthread_mutex_lock(gyro->mutex);
	switch (cmd) {
	case CMD_GYROBOARD_CONNECT:
		if (m->data[0] == 1 && m->data[1] == 2 
		    && m->data[2] == 3 && m->data[3] == 4
		    && m->data[4] == 5 && m->data[5] == 6
		    && m->data[6] == 7) {
			gyro->connected = TRUE;
		} else {
			gyro->connected = FALSE;
		}
		break;
	case CMD_GYROBOARD_GETGYRO:
		gyro->value = (m->data[0] << 24) 
		    | (m->data[1] << 16)
		    | (m->data[2] << 8) 
		    | (m->data[3]);
		break;
	case CMD_GYROBOARD_GETDATA:
		gyro->accel1 = (m->data[0] << 8) | m->data[1];
		gyro->accel2 = (m->data[2] << 8) | m->data[3];
		break;
	default:
		pthread_mutex_unlock(gyro->mutex);
		return FALSE;
	} /* switch */
	pthread_mutex_unlock(gyro->mutex);
	return TRUE;
}

/*----------------------------------------------------------------------*/
/*
 * Returns the current angle of the gyroscope 
 */
double 
canGyroGetAngle(NEO_STR *neo)
{
	CAN_GYRO_STR *gyro = neo->gyro;
	double angle;

	pthread_mutex_lock(gyro->mutex);
	angle = gyro->value * gyro->scale;
	pthread_mutex_unlock(gyro->mutex);
	return angle;
}

/*----------------------------------------------------------------------*/
/*
 * Returns the acceleration data (?)
 */
void
canGyroGetAccel(NEO_STR *neo, double *pAccel1, double *pAccel2)
{
	CAN_GYRO_STR *gyro = neo->gyro;

	pthread_mutex_lock(gyro->mutex);
	*pAccel1 = gyro->accel1/1000.0;
	*pAccel2 = gyro->accel2/1000.0;
	pthread_mutex_unlock(gyro->mutex);
}

/*----------------------------------------------------------------------*/
/* 
 * Display the gyroscope angle and acceleration data on stdout
 */
void
canGyroShow(NEO_STR *neo)
{
	CAN_GYRO_STR *gyro = neo->gyro;
	double a1, a2;

	printf("connected: %s\n", gyro->connected ? "TRUE" : "FALSE");
	printf("angle: %lf\n", canGyroGetAngle(neo));
	canGyroGetAccel(neo, &a1, &a2);
	printf("acceleration: %lf %lf\n", a1, a2);
}

